# -*- coding: utf-8 -*-
"""
Basic scenario tests.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys
import os
from inspect import getsourcefile

from nose.importer import Importer

from aloe import world
from aloe.exceptions import StepDiscoveryError

from aloe.testing import (
    FeatureTest,
    in_directory,
)
from aloe.utils import PY3, TestWrapperIO

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


@in_directory('tests/simple_app')
class SimpleScenarioTest(FeatureTest):
    """
    Test that basic feature running works as expected.
    """

    def test_success(self):
        """
        Test running a simple feature.
        """

        self.assert_feature_success('features/calculator.feature')

    def test_success_zh(self):
        """
        Test running a simple feature in Chinese.
        """

        self.assert_feature_success('features/calculator_zh.feature')

    def test_failure(self):
        """
        Test that a failing feature fails tests.
        """

        self.assert_feature_fail('features/wrong_expectations.feature')

    def step_module_filename(self, module_name):
        """The file name corresponding to the steps module."""

        Importer().importFromDir('.', module_name)
        module = getsourcefile(sys.modules[module_name])
        del sys.modules[module_name]
        return module

    def test_error_message(self):
        """
        Check that the appropriate error messages are printed on failure.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations.feature'

        self.assert_feature_fail(failing_feature, '-n', '1', stream=stream)

        output = stream.getvalue()

        error_header = "FAIL: Fail at adding " + \
            "(features.wrong_expectations: Wrong expectations)"

        self.assertIn(error_header, output)

        feature_stack_frame = """
  File "{feature}", line 11, in Fail at adding
    Then the result should be 40 on the screen
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(feature_stack_frame, output)

        step_file = self.step_module_filename('features.steps')

        step_stack_frame = """
  File "{step_file}", line 62, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_error_message_background(self):
        """
        Check that the appropriate error messages are printed on failure in a
        step background.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations_background.feature'

        self.assert_feature_fail(failing_feature, '-n', '1', stream=stream)

        output = stream.getvalue()

        error_header = "FAIL: Never reached " + \
            "(features.wrong_expectations_background: Wrong expectations)"

        self.assertIn(error_header, output)

        feature_stack_frame = """
  File "{feature}", line 11, in background
    Then the result should be 40 on the screen
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(feature_stack_frame, output)

        step_file = self.step_module_filename('features.steps')

        step_stack_frame = """
  File "{step_file}", line 62, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_failure_zh(self):
        """
        Test that a failing feature in Chinese fails tests.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations_zh.feature'

        self.assert_feature_fail(failing_feature, stream=stream)

        # Check that the appropriate error messages were printed

        output = stream.getvalue()

        error_header = "FAIL: 添加两个数值 " + \
            "(features.wrong_expectations_zh: 不对的预期)"

        self.assertIn(error_header, output)

        if PY3:
            feature_stack_frame = """
  File "{feature}", line 12, in 添加两个数值
    那么结果应该是40
            """.strip().format(feature=os.path.abspath(failing_feature))

            self.assertIn(feature_stack_frame, output)
        else:
            # Cannot use non-ASCII method names in Python 2
            feature_stack_frame = """
  File "{feature}", line 12, in
            """.strip().format(feature=os.path.abspath(failing_feature))

            self.assertIn(feature_stack_frame, output)

            feature_code_line = "那么结果应该是40"
            self.assertIn(feature_code_line, output)

        step_file = self.step_module_filename('features.steps')

        step_stack_frame = """
  File "{step_file}", line 62, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_step_not_found(self):
        """
        Check the behavior when a step is not defined.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/step_not_found.feature'

        self.assert_feature_fail(failing_feature, stream=stream)

        output = stream.getvalue()

        error_header = "NoDefinitionFound: " + \
            "The step r\"When I engage the warp drive\" is not defined"

        self.assertIn(error_header, output)

    def test_step_not_found_zh(self):
        """
        Check the behavior when a step is not defined with a Chinese feature.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/step_not_found_zh.feature'

        self.assert_feature_fail(failing_feature, stream=stream)

        output = stream.getvalue()

        error_header = "NoDefinitionFound: " + \
            "The step r\"当我开曲速引擎\" is not defined"

        self.assertIn(error_header, output)

    def test_background(self):
        """
        Test running a scenario with a background.
        """

        self.assert_feature_success('features/background.feature')

    def test_outlines(self):
        """
        Test a scenario with outlines.
        """

        self.assert_feature_success('features/outlines.feature')

    def test_error_message_outline(self):
        """
        Check that the appropriate error messages are printed on failure in a
        scenario outline.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations.feature'

        self.assert_feature_fail(failing_feature, '-n', '2', stream=stream)

        output = stream.getvalue()

        error_header = "FAIL: Fail repeatedly " + \
            "(features.wrong_expectations: Wrong expectations)"

        self.assertIn(error_header, output)

        # TODO: Why isn't the line indented with 6 spaces like in the file?
        feature_stack_frame = """
  File "{feature}", line 22, in Fail repeatedly: Example 1
    | 50     |
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(feature_stack_frame, output)

        example_stack_frame = """
  File "{feature}", line 18, in Fail repeatedly
    Then the result should be <result> on the screen
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(example_stack_frame, output)

        step_file = self.step_module_filename('features.steps')

        step_stack_frame = """
  File "{step_file}", line 62, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_python_test_skipped(self):
        """
        Test that the Python test does not get picked up.
        """

        self.assert_feature_success('tests')

    def test_non_ascii_files(self):
        """
        Test that features are loaded from a directory also containing a file
        with non-ASCII characters in the name.
        """

        # Nose behavior depends on whether there's __init__.py in the directory
        self.assert_feature_success('features/non_ascii_files')
        self.assert_feature_success('features/non_ascii_files_2')

    def test_scenario_indices(self):
        """
        Test specifying the scenario indices to run.
        """

        feature = 'features/scenario_index.feature'

        # Run a single scenario
        self.assert_feature_success(feature, '-n', '1')
        self.assertEqual(world.all_results, [10])

        # Run multiple scenarios
        # TODO: Handling multi-valued options in Nose is tricky
        self.assert_feature_success(feature, '-n', '1,2')
        self.assertEqual(world.all_results, [10, 20])

        # Run a scenario outline
        self.assert_feature_success(feature, '-n', '3')
        self.assertEqual(world.all_results, [30, 40])

    def test_tags(self):
        """
        Test specifying the tags to run.
        """

        feature_dir = 'features/with_tags'
        feature_one = 'features/with_tags/one.feature'

        # Run scenarios by tag
        self.assert_feature_success(feature_one, '-a', 'hana')
        self.assertEqual(world.all_results, [1, 11, 22])

        # Run scenarios with a tag that's not there
        self.assert_feature_success(feature_one, '-a', 'set')
        self.assertEqual(world.all_results, [])

        # Run features and scenarios with the tag
        self.assert_feature_success(feature_dir, '-a', 'dul')
        self.assertEqual(world.all_results, [2, 13, 20, 200])

        # Specify a tag to exclude
        self.assert_feature_success(feature_one, '-a', '!hana')
        self.assertEqual(world.all_results, [2, 4])

        # Specify more than one tag to run
        self.assert_feature_success(feature_one, '-a', 'hana', '-a', 'dul')
        self.assertEqual(world.all_results, [1, 2, 11, 22])

        # Specify more than one tag to exclude
        self.assert_feature_success(feature_dir, '-a', '!hana,!dul')
        self.assertEqual(world.all_results, [4])


class BadStepsTest(FeatureTest):
    """
    Test loading an application with an error in a step definition file.
    """

    @in_directory('tests/bad_steps_app')
    def test_parent_import_error(self):
        """
        Test the error message when a step definition file cannot be imported
        due to the parent __init__.py file not being found.
        """

        with self.assertRaises(StepDiscoveryError) as raised:
            self.run_features('features/dummy.feature')

        # The file that caused the problem should be visible
        self.assertEqual(
            str(raised.exception),
            "Cannot load step definition file: 'features/steps/__init__.py'"
        )

        # The original exception, with an unhelpful error message. Real cause:
        # features/__init__.py does not exist
        cause = raised.exception.__cause__

        self.assertIsInstance(cause, ImportError)
        # Python 3 has quotes around the module name:
        # No module named 'features'
        self.assertEqual(
            str(cause).replace('\'', ''),
            "No module named features"
        )

    @in_directory('tests/bad_steps_app_2')
    def test_normal_import_error(self):
        """
        Test the error message when a step definition file cannot be imported
        due to an error inside it.
        """

        with self.assertRaises(ValueError):
            self.run_features('features/dummy.feature')



"""
Utilities for tests.
"""

from contextlib import contextmanager


def appender(target, *args):
    """
    A function that appends given arguments and its own to the specified
    list.
    """

    def append(*append_args):
        """Append the earlier specified and given arguments to the target."""
        target.append(args + append_args)

    return append


def before_after(before, after):
    """
    A context manager calling given functions before and after the context,
    respectively, with the same arguments as it is given itself.
    """

    @contextmanager
    def around(*args, **kwargs):
        """Call given functions before and after the context."""
        before(*args, **kwargs)
        yield
        after(*args, **kwargs)

    return around



"""
Aloe integration with `factory_boy`_ to create objects from factories.

Remember when writing BDD tests to describe the behavior you want and not just
use Aloe as a syntax for writing complex tests (that defeats the point of BDD).
Hide the complexity of setting up the objects in your factory or write a
custom step.

To activate these steps import :mod:`aloe.steps.factoryboy` into your
``steps/__init__.py``.

.. _factory_boy: https://factoryboy.readthedocs.org/en/latest/
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import

import re

from aloe import before, step
from aloe.tools import guess_types
from aloe.utils import camel_case_to_spaces


def _run_factory(factory, self, count=None):
    """
    Run a factory using the step definition.
    """

    count = 1 if count is None else int(count)

    if self.table:
        headers = self.table[0]
        rows = tuple(
            dict(zip(headers, row))
            for row in guess_types(self.table[1:])
        )
    else:
        rows = ({},)

    if count != 1 and len(rows) != 1:
        raise ValueError("Cannot specify count and a table together.")

    for row in rows:
        for _ in range(count):
            factory(**row)


def _get_factory_attr(factory, attr):
    """
    Try getting a meta attribute 'attr' from a factory.

    The attribute is looked up as '_attr' on the factory, then as 'attr' on its
    model's meta. The factory's own meta cannot define custom attributes and is
    skipped.

    If the attribute is not found in either place, an AttributeError is raised.
    """

    try:
        return getattr(factory, '_' + attr)
    except AttributeError:
        return getattr(factory._meta.model._meta, attr)  # pylint:disable=protected-access


def step_from_factory(factory):
    """
    Decorator to register a :class:`factory.Factory` as an Aloe step:

        Given/And I have (a/an/`n`) `object(s)`

    An optional table can be passed containing attributes that would be
    passed as `kwargs` to :meth:`factory.Factory.create`.
    Multiple rows or a number of objects can be passed to create more than one
    object. If a number of objects is requested, at most one row can be given,
    passed as `kwargs` to :meth:`factory.Factory.create_batch`.

    The name of the object and its plural are taken from
    :attr:`factory.Factory._meta` if set, or otherwise inferred from the
    class name.

    Example:

    .. code-block:: python

        @step_from_factory
        class RandomUserFactory(factory.Factory):
            '''See Factory Boy docs'''

            class Meta:
                model = models.User

            first_name = factory.Faker('first_name')
            last_name = factory.Faker('last_name')

    .. code-block:: gherkin

        Given I have a random user
        # Then I have created 1 user: Lucy Murray (a random name)

        Given I have 10 random users
        # Then I have created 10 users with different random names

        Given I have random users:
            | first_name | last_name |
            | Danielle   | Madeley   |
            | Alexey     | Kotlyarov |
        # Then I have created 2 users: Danni and Alexey

        Given I have 10 random users:
            | first_name |
            | Joe        |
        # Then I have created 10 users all with the first name Joe
    """

    # Look for verbose_name and verbose_name_plural on the factory or the Meta
    # of the class factory creates; its own Meta cannot define arbitrary
    # attributes.

    try:
        name = _get_factory_attr(factory, 'verbose_name')
    except AttributeError:
        name = camel_case_to_spaces(
            re.sub('Factory$', '', factory.__name__))

    try:
        plural = _get_factory_attr(factory, 'verbose_name_plural')
    except AttributeError:
        plural = name + 's'

    # pylint:disable=unused-variable
    # Functions are declared solely for the decorator side effects
    # (registering them as Lettuce steps and callbacks)

    @step(r'I have an? {name}:?$'.format(name=name))
    @step(r'I have(?: (\d+))? {plural}:?$'.format(plural=plural))
    def run_this_factory(self, count=1):
        """
        Run the factory using the step definition.
        """

        _run_factory(factory, self, count)

    @before.each_example
    def reset_this_factory(self, *args, **kwargs):
        """
        Reset this factory before each example.
        """

        try:
            factory.reset_sequence()
        except (ValueError, AttributeError):
            pass

    return factory



"""
Exception classes
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import super
# pylint:enable=redefined-builtin

from aloe.utils import PY3


class AloeSyntaxError(SyntaxError):
    """A syntax error in a feature file."""
    def __init__(self, filename, string):
        self.filename = filename
        self.string = string

        msg = "Syntax error at: {filename}\n{string}".format(
            filename=filename, string=string)

        super().__init__(msg)


class StepLoadingError(Exception):
    """A step definition cannot be registered."""
    pass


class StepDiscoveryError(Exception):
    """A step definition file cannot be imported."""
    pass


class NoDefinitionFound(Exception):
    """
    Exception raised when there is no suitable step definition for a step.
    """

    def __init__(self, step):
        self.step = step
        super().__init__('The step r"%s" is not defined' % self.step.sentence)


# Step definitions can contain Unicode, and that will propagate to the
# exception message. On Python 2, this makes handling such exceptions tricky,
# so force encoding the message as Unicode.
if not PY3:
    NoDefinitionFound.__str__ = lambda self: self.message.encode('utf-8')


def undefined_step(step, *args, **kwargs):
    """
    A fallback step used when no suitable step definition was found.

    Raises a NoDefinitionFound exception when the scenario is run.
    """
    raise NoDefinitionFound(step)



"""
Test running Aloe in a separate process, as the real users would.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import subprocess
import unittest

from aloe.testing import in_directory


ROOT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
MAIN = os.path.join(ROOT_PATH, 'aloe', '__init__.py')


@in_directory('tests/simple_app')
class SimpleIntegrationTest(unittest.TestCase):
    """
    Test that basic feature running works as expected.
    """

    def test_success(self):
        """
        Test running a simple feature.
        """

        exitcode, _ = self.run_feature('features/calculator.feature')
        self.assertEqual(exitcode, 0, "Feature run successfully.")

    def test_failure(self):
        """
        Test running a failing feature.
        """

        exitcode, _ = self.run_feature('features/wrong_expectations.feature')
        self.assertNotEqual(exitcode, 0, "Feature run failed.")

    def run_feature(self, *args):
        """
        Run a feature and return the (exitcode, output) tuple.
        """

        # Ensure Aloe itself is on the path
        env = os.environ.copy()
        env['PYTHONPATH'] = ROOT_PATH

        try:
            output = subprocess.check_output(
                (MAIN,) + args,
                stderr=subprocess.STDOUT,
                env=env,
            )
            return 0, output
        except subprocess.CalledProcessError as ex:
            return ex.returncode, ex.output



"""
Test feature loading across a symlink.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os

from aloe.testing import (
    FeatureTest,
    in_directory,
)


@in_directory('tests/symlink_follow_app/working_path')
class SymlinkLoadingTest(FeatureTest):
    """
    Test that symlink feature is discovered and evaluated.
    """

    def test_symlinked_feature(self):
        """
        Test that we can discover feature in a symlinked module.
        """

        result = self.assert_feature_success()
        self.assertEqual(
            result.tests_run,
            [os.path.abspath('common_app/features/symlinked.feature')]
        )
